==> 0.0.7.tar.gz: lib/src/firestore_repository.dart <==
import 'package:cloud_firestore/cloud_firestore.dart';
import 'package:firestore_repository_helper/src/firestore_object_extensions.dart';
import 'package:firestore_repository_helper/src/t_snapshot.dart';
import 'package:future_helper/future_helper.dart';
import 'package:resource_result/resource_result.dart';
import 'package:serializer_interface/serializer_interface.dart';

import 'firestore_object.dart';
import 'no_data_failure.dart';
import 'snapshot_serialization_exception.dart';

typedef QueryApplier<T extends FirestoreObject<T>> = Query<T> Function(
    CollectionReference<T> collection);

abstract class FirestoreRepository<T extends FirestoreObject<T>> {
  String get collectionPath;

  JsonSerializer<T> get serializer;

  /// returns [Failure] if there was an error.
  /// returns [NoDataFailure] if documents do not exist
  Future<Resource<List<T>>> readAll();

  Future<Resource<List<T>>> query(QueryApplier<T> applyQuery);

  /// returns [Failure] if there was an error.
  /// returns [NoDataFailure] if document does not exist
  Future<Resource<T>> readDocument(String id);

  /// returns [Failure] if there was an error.
  /// returns [NoDataFailure] if document does not exist
  Future<Resource<TSnapshot<T>>> readDocumentWithMetadata(String id);

  /// throws [Exception] if there was an error.
  Future createDocument(T data);

  /// throws [Exception] if there was an error.
  Future updateDocument(T data, {bool createIfDoesNotExist = false});

  /// throws [Exception] if there was an error.
  Future updateDocumentFields(String id, Map<String, dynamic> fields);

  /// throws [Exception] if there was an error.
  Future deleteDocument(String id);

  /// streams [Failure] if there was an error.
  /// streams [NoDataFailure] if document does not exist
  Stream<Resource<T>> listenToDocument(String id);

  /// streams [Failure] if there was an error.
  /// streams [NoDataFailure] if document does not exist
  Stream<Resource<TSnapshot<T>>> listenToDocumentWithMetadata(String id);

  /// streams [Failure] if there was an error.
  /// streams [NoDataFailure] if no documents exist
  Stream<Resource<List<T>>> listenToAll();

  /// streams [Failure] if there was an error.
  /// streams [NoDataFailure] if no documents exist
  Stream<Resource<List<T>>> listenToQuery(QueryApplier<T> applyQuery);
}

abstract class FirestoreRepositoryImpl<T extends FirestoreObject<T>>
    implements FirestoreRepository<T> {
  CollectionReference<T> _getCollectionRef() {
    // for converter see: https://firebase.flutter.dev/docs/firestore/usage/#typing-collectionreference-and-documentreference
    return FirebaseFirestore.instance
        .collection(collectionPath)
        .withConverter<T>(
            fromFirestore: _fromSnapshotConverter,
            toFirestore: _toSnapShotConverter);
  }

  T _fromSnapshotConverter(DocumentSnapshot<Map<String, dynamic>> snapshot,
      SnapshotOptions? options) {
    try {
      return serializer.fromJson(snapshot.data()!).withId(snapshot.id);
    } catch (e) {
      throw SerializationException.fromJson(snapshot.data());
    }
  }

  Map<String, dynamic> _toSnapShotConverter(
      T firestoreObj, SetOptions? options) {
    try {
      return serializer.toJson(firestoreObj);
    } catch (e) {
      throw SerializationException.toJson(firestoreObj);
    }
  }

  @override
  Future deleteDocument(String id) => _getCollectionRef().doc(id).delete();

  @override
  Future<Resource<List<T>>> readAll() =>
      _getCollectionRef().get().map(_querySnapshotToResourceList);

  @override
  Future<Resource<List<T>>> query(QueryApplier<T> applyQuery) =>
      applyQuery(_getCollectionRef()).get().map(_querySnapshotToResourceList);

  Resource<List<T>> _querySnapshotToResourceList(
      QuerySnapshot<T> querySnapshot) {
    try {
      final objects =
          querySnapshot.docs.map((snapshot) => snapshot.data()).toList();
      return Success(objects);
    } catch (e) {
      return Failure.fromMessage(
          "Failed to convert query snapshot to list ${e.toString()}");
    }
  }

  @override
  Future<Resource<T>> readDocument(String id) =>
      _getDocuRef(id).get().map(_fromSnapshot);

  @override
  Future<Resource<TSnapshot<T>>> readDocumentWithMetadata(String id) =>
      _getDocuRef(id).get().map(_fromSnapshotWithMetadata);

  DocumentReference<T> _getDocuRef(String id) => _getCollectionRef().doc(id);

  @override
  Future createDocument(T data) => _getDocuRef(data.id).set(data);

  @override
  Future updateDocument(T data, {bool createIfDoesNotExist = false}) =>
      createIfDoesNotExist
          ? createDocument(data)
          : updateDocumentFields(data.id, serializer.toJson(data));

  @override
  Future updateDocumentFields(String id, Map<String, dynamic> fields) =>
      _getDocuRef(id).update(fields);

  @override
  Stream<Resource<T>> listenToDocument(String id) =>
      _getDocuRef(id).snapshots().map(_fromSnapshot);

  @override
  Stream<Resource<TSnapshot<T>>> listenToDocumentWithMetadata(String id) =>
      _getDocuRef(id).snapshots().map(_fromSnapshotWithMetadata);

  @override
  Stream<Resource<List<T>>> listenToAll() =>
      _getCollectionRef().snapshots().map(_querySnapshotToResourceList);

  @override
  Stream<Resource<List<T>>> listenToQuery(QueryApplier<T> applyQuery) =>
      applyQuery(_getCollectionRef())
          .snapshots()
          .map(_querySnapshotToResourceList);

  Resource<T> _fromSnapshot(DocumentSnapshot<T> snapshot) {
    if (!snapshot.exists || snapshot.data() == null) {
      return NoDataFailure(
          Error('No document with id "${snapshot.id}" with data found.'));
    }
    return Success(snapshot.data()!);
  }

  Resource<TSnapshot<T>> _fromSnapshotWithMetadata(
      DocumentSnapshot<T> snapshot) {
    if (!snapshot.exists || snapshot.data() == null) {
      return NoDataFailure(
          Error('No document with id "${snapshot.id}" with data found.'));
    }
    final metadata = MetaData(
        hasPendingWrites: snapshot.metadata.hasPendingWrites,
        isFromCache: snapshot.metadata.isFromCache);
    return Success(TSnapshot(snapshot.data()!, metaData: metadata));
  }
}

==> 0.0.9.tar.gz: lib/src/firestore_repository.dart <==
import 'package:cloud_firestore/cloud_firestore.dart';
import 'package:firestore_repository_helper/src/firestore_object_extensions.dart';
import 'package:firestore_repository_helper/src/t_snapshot.dart';
import 'package:future_helper/future_helper.dart';
import 'package:resource_result/resource_result.dart';
import 'package:serializer_interface/serializer_interface.dart';

import 'firestore_object.dart';
import 'no_data_failure.dart';
import 'snapshot_serialization_exception.dart';

typedef QueryApplier<T extends FirestoreObject<T>> = Query<T> Function(
    CollectionReference<T> collection);

abstract class FirestoreRepository<T extends FirestoreObject<T>> {
  String get collectionPath;

  JsonSerializer<T> get serializer;

  /// returns [Failure] if there was an error.
  /// returns [NoDataFailure] if documents do not exist
  Future<Resource<List<T>>> readAll();

  Future<Resource<List<T>>> query(QueryApplier<T> applyQuery);

  /// returns [Failure] if there was an error.
  /// returns [NoDataFailure] if document does not exist
  Future<Resource<T>> readDocument(String id);

  /// returns [Failure] if there was an error.
  /// returns [NoDataFailure] if document does not exist
  Future<Resource<TSnapshot<T>>> readDocumentWithMetadata(String id);

  /// throws [Exception] if there was an error.
  Future createDocument(T data);

  /// throws [Exception] if there was an error.
  Future updateDocument(T data, {bool createIfDoesNotExist = false});

  /// throws [Exception] if there was an error.
  Future updateDocumentFields(String id, Map<String, dynamic> fields);

  /// throws [Exception] if there was an error.
  Future deleteDocument(String id);

  /// streams [Failure] if there was an error.
  /// streams [NoDataFailure] if document does not exist
  Stream<Resource<T>> listenToDocument(String id);

  /// streams [Failure] if there was an error.
  /// streams [NoDataFailure] if document does not exist
  Stream<Resource<TSnapshot<T>>> listenToDocumentWithMetadata(String id);

  /// streams [Failure] if there was an error.
  /// streams [NoDataFailure] if no documents exist
  Stream<Resource<List<T>>> listenToAll();

  /// streams [Failure] if there was an error.
  /// streams [NoDataFailure] if no documents exist
  Stream<Resource<List<T>>> listenToQuery(QueryApplier<T> applyQuery);
}

abstract class FirestoreRepositoryImpl<T extends FirestoreObject<T>>
    implements FirestoreRepository<T> {
  CollectionReference<T> _getCollectionRef() {
    // for converter see: https://firebase.flutter.dev/docs/firestore/usage/#typing-collectionreference-and-documentreference
    return FirebaseFirestore.instance
        .collection(collectionPath)
        .withConverter<T>(
            fromFirestore: _fromSnapshotConverter,
            toFirestore: _toSnapShotConverter);
  }

  T _fromSnapshotConverter(DocumentSnapshot<Map<String, dynamic>> snapshot,
      SnapshotOptions? options) {
    try {
      return serializer.fromJson(snapshot.data()!).withId(snapshot.id);
    } catch (e) {
      throw SerializationException.fromJson(snapshot.data());
    }
  }

  Map<String, dynamic> _toSnapShotConverter(
      T firestoreObj, SetOptions? options) {
    try {
      return serializer.toJson(firestoreObj);
    } catch (e) {
      throw SerializationException.toJson(firestoreObj);
    }
  }

  @override
  Future deleteDocument(String id) => _getCollectionRef().doc(id).delete();

  @override
  Future<Resource<List<T>>> readAll() =>
      _getCollectionRef().get().map(_querySnapshotToResourceList);

  @override
  Future<Resource<List<T>>> query(QueryApplier<T> applyQuery) =>
      applyQuery(_getCollectionRef()).get().map(_querySnapshotToResourceList);

  Resource<List<T>> _querySnapshotToResourceList(
      QuerySnapshot<T> querySnapshot) {
    try {
      final objects =
          querySnapshot.docs.map((snapshot) => snapshot.data()).toList();
      return Success(objects);
    } catch (e) {
      return Failure.fromMessage(
          "Failed to convert query snapshot to list ${e.toString()}");
    }
  }

  @override
  Future<Resource<T>> readDocument(String id) =>
      _getDocuRef(id).get().map(_fromSnapshot);

  @override
  Future<Resource<TSnapshot<T>>> readDocumentWithMetadata(String id) =>
      _getDocuRef(id).get().map(_fromSnapshotWithMetadata);

  DocumentReference<T> _getDocuRef(String id) => _getCollectionRef().doc(id);

  @override
  Future createDocument(T data) => _getDocuRef(data.id).set(data);

  @override
  Future updateDocument(T data, {bool createIfDoesNotExist = false}) =>
      createIfDoesNotExist
          ? createDocument(data)
          : updateDocumentFields(data.id, serializer.toJson(data));

  @override
  Future updateDocumentFields(String id, Map<String, dynamic> fields) =>
      _getDocuRef(id).update(fields);

  @override
  Stream<Resource<T>> listenToDocument(String id,
          {bool includeMetadataChanges = false}) =>
      _getDocuRef(id)
          .snapshots(includeMetadataChanges: includeMetadataChanges)
          .map(_fromSnapshot);

  @override
  Stream<Resource<TSnapshot<T>>> listenToDocumentWithMetadata(String id,
          {bool includeMetadataChanges = false}) =>
      _getDocuRef(id)
          .snapshots(includeMetadataChanges: includeMetadataChanges)
          .map(_fromSnapshotWithMetadata);

  @override
  Stream<Resource<List<T>>> listenToAll(
          {bool includeMetadataChanges = false}) =>
      _getCollectionRef()
          .snapshots(includeMetadataChanges: includeMetadataChanges)
          .map(_querySnapshotToResourceList);

  @override
  Stream<Resource<List<T>>> listenToQuery(QueryApplier<T> applyQuery,
          {bool includeMetadataChanges = false}) =>
      applyQuery(_getCollectionRef())
          .snapshots(includeMetadataChanges: includeMetadataChanges)
          .map(_querySnapshotToResourceList);

  Resource<T> _fromSnapshot(DocumentSnapshot<T> snapshot) {
    if (!snapshot.exists || snapshot.data() == null) {
      return NoDataFailure(
          Error('No document with id "${snapshot.id}" with data found.'));
    }
    return Success(snapshot.data()!);
  }

  Resource<TSnapshot<T>> _fromSnapshotWithMetadata(
      DocumentSnapshot<T> snapshot) {
    if (!snapshot.exists || snapshot.data() == null) {
      return NoDataFailure(
          Error('No document with id "${snapshot.id}" with data found.'));
    }
    final metadata = MetaData(
        hasPendingWrites: snapshot.metadata.hasPendingWrites,
        isFromCache: snapshot.metadata.isFromCache);
    return Success(TSnapshot(snapshot.data()!, metaData: metadata));
  }
}
